linktime-proc-macro 0.0.1

proc-macro support for the linktime crates
docs.rs failed to build linktime-proc-macro-0.0.1
Please check the build logs for more information.
See Builds for ideas on how to fix a failed build, or Metadata for how to configure docs.rs builds.
If you believe this is docs.rs' fault, open an issue.
Visit the last successful build: linktime-proc-macro-0.1.0

Support crate only

This crate is not intended to be used directly and does not have any public exports.

See ctor and dtor on crates.io for the main crates.